/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.esb;


import org.apache.axis2.AxisFault;
import org.apache.axis2.description.Parameter;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.context.ConfigurationContextFactory;
import org.apache.axis2.engine.ListenerManager;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.derby.drda.NetworkServerControl;
import org.apache.synapse.config.SynapseConfiguration;
import org.apache.synapse.registry.Registry;
import org.wso2.adminui.UIProcessingException;
import org.wso2.adminui.UIProcessor;
import org.wso2.esb.util.XmlConfigurationFactory;
import org.wso2.esb.util.JettyWebConfiguration;
import org.wso2.esb.registry.ESBRegistry;
import org.wso2.utils.ServerException;

import java.io.File;
import java.net.UnknownHostException;
import java.util.Hashtable;
import java.util.Properties;

/*
 * The Service bus manager
 */
public class ServiceBusManager {

    private static Log log = LogFactory.getLog(ServiceBusManager.class);

    public static int httpPort  = 8080;
    public static int httpsPort = 9443;

    private ServiceBusConfiguration esbConfiguration;
    public ConfigurationContext configurationContext;
    private NetworkServerControl networkServerControl;

    public ServiceBusManager() {
        this(ServiceBusConfiguration.getServiceBusInstance());
    }

    public ServiceBusManager(ServiceBusConfiguration esbConfiguration) {
        this.esbConfiguration = esbConfiguration;
    }

    /**
     * This implementation does not validaye the user against a database. In future it
     * is envisioned that the ESB will reuse the common authentication components shared
     * between WSAS and ESB
     * @param username
     * @param password
     * @return true if the user is authenticated successfully
     */
    public boolean authenticate(String username, String password) {

        String esb_username = esbConfiguration.getFirstProperty("DefaultAdmin.Username");
        String esb_password = esbConfiguration.getFirstProperty("DefaultAdmin.Password");

        return username.equals(esb_username) &&
                password.equals(esb_password);
    }

    /**
     * Starts the Derby database instance in network mode for multi user access
     * Waits till the server can be pinged, upto a maximum of 10 seconds
     * @throws ServiceBusException
     */
    public void startDataBaseServer() throws ServiceBusException {

        String host = null;
        try {
            networkServerControl = new NetworkServerControl();
            networkServerControl.start(null);

            for (int i=0; i<10; i++) {
                Thread.sleep(1000);
                try {
                    networkServerControl.ping();
                    Properties properties = networkServerControl.getCurrentProperties();
                    host = properties.getProperty("derby.drda.host");
                    int port = Integer.parseInt(properties.getProperty("derby.drda.portNumber"));
                    log.info("Database server started on " + host + " over port " + port);
                    return;

                } catch (Exception ignore) {}
            }
            handleFatal("Database server failed to start");

        } catch (UnknownHostException e) {
            handleException("Unknown host : " + host, e);
        } catch (Exception e) {
            handleException("Error starting the database server", e);
        }
    }

    /**
     * Requests the Derby DB to stop itself
     * @throws ServiceBusException
     */
    public void stopDataBaseServer() throws ServiceBusException {
        if (networkServerControl != null) {
            try {
                networkServerControl.shutdown();
                log.info("Database server shutting down");
            } catch (Exception e) {
                throw new ServiceBusException(e);
            }
        }
    }

    public void start() throws AxisFault, ServiceBusException {

        log.debug("Starting the ServiceBusManager");

        initSystemProperties();
        startDataBaseServer();

        //axis2.xml ; axis2.repo and synapse.xml will inject as System properties
        configurationContext = ConfigurationContextFactory.
            createConfigurationContextFromFileSystem(null, null);

        // configuration has now been loaded from the configuration file. now set ESB default values
        // for unspecified entities.
        setESBDefaults(configurationContext);

        initListenerManager();

        // intializing configs related to JettyServer
        try {
            XmlConfigurationFactory.init(Constants.ESB_WEB_XML_KEY,
                "conf/server-web.xml", Constants.ESB_XML_NAMESPACE);
        } catch (ServerException e) {
            handleException("Error loading server-web.xml", e);
        }
        JettyWebConfiguration.getInstance().setXmlFileKey(Constants.ESB_WEB_XML_KEY);
        JettyWebConfiguration.getInstance().init();

        // Create the index.html page and save it in the ConfigContext
        generateWelcomePage();

    }

    /**
     * Assign deafult values for synapse configuration which are specific to ESB.
     *
     * @param configContext ConfigurationContext
     */
    private void setESBDefaults(ConfigurationContext configContext) {

        Parameter configParameter = configurationContext.
                getAxisConfiguration().getParameter(org.apache.synapse.Constants.SYNAPSE_CONFIG);
        Object o = configParameter.getValue();

        if (o != null && o instanceof SynapseConfiguration) {
            SynapseConfiguration synConfig = (SynapseConfiguration) o;

            // set default registry to ESB registry
            if (synConfig.getRegistry() == null) {
                Registry registry = new ESBRegistry();
                registry.addConfigProperty("root", "file:registry");
                registry.addConfigProperty("cachableDuration", "15000");

                synConfig.setRegistry(registry);
            }
        }
    }

    /**
     * Generate the index.html page and cache it in the ConfigurationContext
     */
    private void generateWelcomePage() throws ServiceBusException {

        Hashtable fileContents = new Hashtable();
        try {
            UIProcessor.createPages(
                JettyWebConfiguration.getInstance().adminWebResourceBase, null, fileContents);
        } catch (UIProcessingException e) {
            handleException("Error generating welcome page", e);
        }
        configurationContext.setProperty(Constants.GENERATED_PAGES, fileContents);
    }

    private void initListenerManager() {
        ListenerManager listenerManager = new ListenerManager();
        listenerManager.init(configurationContext);
        listenerManager.start();
        ListenerManager.defaultConfigurationContext = configurationContext;
    }

    private void initSystemProperties() {
        String axis2Xml = System.getProperty("axis2.xml");
        String esbHome  = System.getProperty("esb.home"); // this has to be provided

        if (axis2Xml == null) {
            System.setProperty("axis2.xml",
                esbHome + File.separator + "conf" + File.separator + "axis2.xml");
        }

        String esbXml = System.getProperty("esb.xml");
        if (esbXml == null) {
            System.setProperty("esb.xml", ServiceBusConfiguration.configurationXMLLocation);
        }

        String synapseXml = System.getProperty("synapse.xml");
        if (synapseXml == null) {
            System.setProperty("synapse.xml",
                esbHome + File.separator + "conf" + File.separator + "synapse.xml");
        }

        System.setProperty("derby.system.home",esbHome + File.separator + "conf");
    }

    private static void handleFatal(String msg) throws ServiceBusException {
        log.fatal(msg);
        throw new ServiceBusException(msg);
    }

    private static void handleException(String msg, Exception e) throws ServiceBusException {
        log.error(msg, e);
        throw new ServiceBusException(msg, e);
    }

}
