/*
* Copyright 2004,2005 The Apache Software Foundation.
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/
package org.wso2.esb.persistence.dao;

import org.apache.log4j.Logger;
import org.hibernate.*;
import org.hibernate.exception.ConstraintViolationException;
import org.wso2.esb.persistence.dataobject.BaseDO;
import org.wso2.esb.util.HibernateConfig;

import java.util.Date;
import java.util.List;

/**
 * The base DataAccessObject from which each DAO of the ESB is extended
 */
public abstract class BaseDAO {

    protected Logger log = Logger.getLogger(BaseDAO.class);

    /**
     * The HibernateConfiguration
     */
    protected HibernateConfig hbConfig;

    public BaseDAO(HibernateConfig hbConfig) {
        this.hbConfig = hbConfig;
    }

    /**
     * To save a entity
     *
     * @param trasientInstance
     * @return The entity ID
     */
    public Long create(BaseDO trasientInstance) {
        Session session = hbConfig.currentSession();
        Transaction tx = session.beginTransaction();
        try {
            trasientInstance.setLastUpdatedTime(new Date());
            session.persist(trasientInstance);
            session.flush();
            tx.commit();
        } catch (ConstraintViolationException e) {
            String msg = "Trying to create duplicate entity : " + trasientInstance;
            log.warn(msg, e);
        } catch (HibernateException e) {
            tx.rollback();
            handleException("Cannot create entity " + trasientInstance, e);
        } finally {
            hbConfig.closeSession();
        }

        return trasientInstance.getId();
    }

    /**
     * To create a new entity or update already existed entity
     *
     * @param instance
     * @return The entity ID
     */
    public Long createOrUpdate(BaseDO instance) {
        Session session = hbConfig.currentSession();
        Transaction tx = session.beginTransaction();
        try {
            instance.setLastUpdatedTime(new Date());
            session.saveOrUpdate(instance);
            session.flush();
            tx.commit();
        } catch (ConstraintViolationException e) {
            String msg = "Trying to create duplicate entity " + instance;
            log.warn(msg, e);
        } catch (HibernateException e) {
            tx.rollback();
            handleException("Cannot create entity " + instance, e);
        } finally {
            hbConfig.closeSession();
        }

        return instance.getId();
    }

    /**
     * To update a already existed entity
     *
     * @param abstractDO
     */
    public void update(BaseDO abstractDO) {
        Session session = hbConfig.currentSession();
        Transaction tx = session.beginTransaction();
        try {
            abstractDO.setLastUpdatedTime(new Date());
            session.merge(abstractDO);
            session.flush();
            tx.commit();
        } catch (HibernateException e) {
            tx.rollback();
            handleException("Unable to update " + abstractDO, e);
        } finally {
            hbConfig.closeSession();
        }
    }

    /**
     * To delete a entity
     *
     * @param abstractDO
     */
    public void delete(BaseDO abstractDO) {
        Session session = hbConfig.currentSession();
        Transaction tx = session.beginTransaction();
        try {
            session.delete(abstractDO);
            session.flush();
            tx.commit();
        } catch (HibernateException e) {
            tx.rollback();
            handleException("Unable to delete " + abstractDO, e);
        } finally {
            hbConfig.closeSession();
        }
    }

    /**
     * To Retrieves entities as a List
     *
     * @param hqlQuery
     * @return The resultset
     */
    public List select(String hqlQuery) {
        Session session = hbConfig.currentSession();
        Transaction tx = session.beginTransaction();
        Query query;
        List result = null;
        try {
            query = session.createQuery(hqlQuery);
            result = query.list();
        } catch (HibernateException e) {
            tx.rollback();
            handleException("Error executing HQL query : " + hqlQuery, e);
        } finally {
            hbConfig.closeSession();
        }
        return result;
    }

    /**
     * To execute a query
     *
     * @param hqlQuery
     * @return The resultset
     */
    public int execute(String hqlQuery) {
        Session session = hbConfig.currentSession();
        Transaction tx = session.beginTransaction();
        Query query;
        int result = -1;
        try {
            query = session.createQuery(hqlQuery);
            result = query.executeUpdate();
        } catch (HibernateException e) {
            tx.rollback();
            handleException("Error executing HQL query : " + hqlQuery, e);
        } finally {
            hbConfig.closeSession();
        }
        return result;
    }

    /**
     * To Retrieves one unique entity as a Object
     *
     * @param hqlQuery
     * @return The entity
     */
    public Object selectUnique(String hqlQuery) {

        Session session = hbConfig.currentSession();
        Transaction tx = session.beginTransaction();
        Query query;
        Object result = null;
        try {
            query = session.createQuery(hqlQuery);
            result = query.uniqueResult();
        } catch (HibernateException e) {
            tx.rollback();
            handleException("Error executing HQL query : " + hqlQuery, e);
        } finally {
            hbConfig.closeSession();
        }
        return result;
    }

    /**
     * To Retrieves entities as a List using SQL
     *
     * @param sqlQuery
     * @param aClass
     * @return The resultset
     */
    public List selectBySQL(String sqlQuery, Class aClass) {
        Session session = hbConfig.currentSession();
        List result = null;

        Transaction tx = session.beginTransaction();
        try {
            SQLQuery query = session.createSQLQuery(sqlQuery).addEntity(aClass);
            result = query.list();

        } catch (HibernateException e) {
            tx.rollback();
            handleException("Error executing SQL : " + sqlQuery, e);
        } finally {
            if (session != null) {
                try {
                    hbConfig.closeSession();
                } catch (HibernateException ignore) {
                    log.warn(ignore);
                }
            }
        }
        return result;
    }

    private void handleException(String msg, Exception e) {
        log.error(msg, e);
        throw new RuntimeException(msg, e);
    }
}

