/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *   * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 */

package org.wso2.esb.registry;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.axiom.om.OMNode;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.impl.builder.StAXOMBuilder;
import org.apache.synapse.registry.RegistryEntry;
import org.apache.synapse.registry.AbstractRegistry;
import org.apache.synapse.registry.Registry;
import org.apache.synapse.SynapseException;
import org.wso2.esb.Constants;
import org.wso2.esb.persistence.dataobject.RegistryEntryDO;
import org.wso2.esb.persistence.PersistenceManager;

import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.SSLSession;
import javax.net.ssl.HttpsURLConnection;
import java.net.*;
import java.io.*;
import java.util.ArrayList;

/**
 * Registry implementation for ESB. This assumes that registry resources can be accessed using file
 * system or http. Registry meta data are accessed using web services.
 */
public class ESBRegistry extends AbstractRegistry implements Registry {

    public static final int FILE = 100;
    public static final int HTTP = 101;
    public static final int HTTPS = 102;
    public static final String URL_SEPARATOR = "/";
    public static final char URL_SEPARATOR_CHAR = '/';

    private static final int DELETE_RETRY_SLEEP_TIME = 10;
    private static final long DEFAULT_CACHABLE_DURATION = 60000;
    private static final Log log = LogFactory.getLog(ESBRegistry.class);

    private static final int MAX_KEYS = 200;

    /**
     * File system path corresponding to the file url path. This is a system depending path
     * used for accessing resources as files.
     */
    private String localRegistry = null;

    /**
     * URL of the WS endpoind to get data about registry entries.
     */
    private String metaDataService = null;

    /**
     * Specifies whether the registry is in the local host or a remote registry.
     * Local host means the same computer as ESB is running.
     */
    private int registryType = Constants.LOCAL_HOST_REGISTRY;

    /**
     * Contains the protocol for the registry. Allowd values are FILE, HTTP and HTTPS.
     */
    private int registryProtocol = FILE;

    public OMNode lookup(String key) {

        log.info("==> Repository fetch of resource with key : " + key);

        URLConnection urlc = null;
        try {
            URL url = new URL(getRoot() + key);
            urlc = url.openConnection();
            urlc.connect();
        } catch (IOException e) {
            return null;
        }

        try {
            XMLStreamReader parser = XMLInputFactory.newInstance().
                    createXMLStreamReader(urlc.getInputStream());
            StAXOMBuilder builder = new StAXOMBuilder(parser);
            return builder.getDocumentElement();

        } catch (MalformedURLException e) {
            handleException("Invalid URL reference " + getRoot() + key, e);
        } catch (IOException e) {
            handleException("IO Error reading from URL " + getRoot() + key, e);
        } catch (XMLStreamException e) {
            handleException("XML Error reading from URL " + getRoot() + key, e);
        }
        return null;
    }

    public RegistryEntry getRegistryEntry(String key) {

        // get information from the actual resource
        ESBRegistryEntry entry = new ESBRegistryEntry();

        try {
            URL url = new URL(getRoot() + key);
            URLConnection urlc = url.openConnection();

            entry.setKey(key);
            entry.setName(url.getFile());
            entry.setType(Constants.file);

            entry.setDescription("Resource at : " + url.toString());
            entry.setLastModified(urlc.getLastModified());
            entry.setVersion(urlc.getLastModified());
            if (urlc.getExpiration() > 0) {
                entry.setCachableDuration(
                        urlc.getExpiration() - System.currentTimeMillis());
            } else {
                entry.setCachableDuration(getCachableDuration());
            }

        } catch (MalformedURLException e) {
            handleException("Invalid URL reference " + getRoot() + key, e);
        } catch (IOException e) {
            handleException("IO Error reading from URL " + getRoot() + key, e);
        }

        // get information from the database
        PersistenceManager persistenceManager = new PersistenceManager();
        RegistryEntryDO registryEntryDO = persistenceManager.getRegistryEntry(key);
        
        if (registryEntryDO != null) {

            if (registryEntryDO.getExpiryTime() != null) {
                entry.setCachableDuration(registryEntryDO.getExpiryTime().longValue());
            } else {
                entry.setCachableDuration(0);
            }
        }

        return entry;
    }

    /**
     * Updates the metadata of the given registry entry
     *
     * @param entry RegistryEntry containing the new metadata
     */
    public void updateRegistryEntry(RegistryEntry entry) {

        RegistryEntryDO registryEntryDO = new RegistryEntryDO();
        registryEntryDO.setRegistryKey(entry.getKey());
        registryEntryDO.setExpiryTime(new Long(entry.getCachableDuration()));

        PersistenceManager persistenceManager = new PersistenceManager();
        persistenceManager.saveOrUpdateRegistryEntry(registryEntryDO);
    }

    /**
     * Updates the registry resource pointed by the given key.
     *
     * @param key Key of the resource to be updated
     * @param value New value of the resource
     * @throws Exception
     */
    public void updateResource(String key, OMElement value) throws Exception {

        if (registryType == Constants.LOCAL_HOST_REGISTRY) {
            File file = new File(localRegistry + getSystemDependentPath(key));
            if (file.exists()) {
                BufferedWriter writer = null;
                try {
                    writer = new BufferedWriter(new FileWriter(file));
                    writer.write(value.toString());
                    writer.flush();
                } catch (IOException e) {
                    throw new RuntimeException("Couldn't write to registry entry: " + key);
                } finally {
                    if (writer != null) {
                        writer.close();
                    }
                }

            }
        } else {
            throw new RuntimeException("Remote registry is not supported.");
        }
    }

    /**
     * Adds a new resource to the registry.
     *
     * @param parentName Key of the parent of the new resource
     * @param resourceName Name of the new resource
     * @param isLeaf Specifies whether the new resource is a leaf or not. In a file system based
     * registry, leaf is a file and non-leaf is a folder.
     * @throws Exception
     */
    public void addResource(String parentName,
                            String resourceName, boolean isLeaf) throws Exception {

        if (registryType == Constants.LOCAL_HOST_REGISTRY) {

            if (isLeaf) {
                createFile(getSystemDependentPath(parentName), resourceName);
            } else {
                createFolder(getSystemDependentPath(parentName), resourceName);
            }
        }
    }

    /**
     * Removes the registry resource identified by the given key. If the key points to a directory,
     * all its subdirectories and files in those directoris will be deleted. All the database
     * entries for deleted registry resources will be removed.
     *
     * @param key
     */
    public void removeResource(String key) {

        if (registryType == Constants.LOCAL_HOST_REGISTRY) {
            File resource = new File(localRegistry + getSystemDependentPath(key));
            if (resource.exists()) {

                if (resource.isFile()) {
                    deleteFile(resource);
                } else if (resource.isDirectory()) {
                    deleteDirectory(resource);
                }

            } else {
                throw new SynapseException("Parent folder: " + key + " does not exists.");
            }
        }
    }

    /**
     * Returns the registry key in URL style (with "/" as the separator) for a given file.
     *
     * @param file File to get the key
     * @return Registry key
     */
    private String getRegsitryKey(File file) {

        String path = file.getAbsolutePath();
        String rootPath = new File(localRegistry).getAbsolutePath();
        return getURLPath(path.substring(rootPath.length() + 1)); // we have remove the preceeding "/"
    }

    private void deleteFile(File file) {

        boolean success = file.delete();
        if (!success) {
            // try with this work around to overcome a known bug in windows
            // work around:
            // run garbage collector and sleep for some time and delete
            // if still didn't delete,
            // rename file to a temp file in "temp" dir and mark it to delete on exist
            
            System.gc();
            try {
                Thread.sleep(DELETE_RETRY_SLEEP_TIME);
            } catch (InterruptedException e) {
                // ignore the exception
            }

            success = file.delete();
            if (!success) {
                int suffix = 1;
                File renamedFile = null;

                File tempDir = new File("temp");
                if (!tempDir.exists()) {
                    tempDir.mkdir();
                }

                do {
                    String changedName = "d" + suffix + file.getName();
                    renamedFile = new File(tempDir, changedName);
                    suffix++;
                } while (renamedFile.exists());

                if (file.renameTo(renamedFile)) {
                    success = true;
                    renamedFile.deleteOnExit();
                } else {
                    throw new RuntimeException("Cannot delete the resource: " + file.getName());
                }
            }
        }

        if (success) {
            PersistenceManager persistenceManager = new PersistenceManager();
            persistenceManager.deleteRegistryEntry(getRegsitryKey(file));
        }
    }

    private void deleteDirectory(File dir) {

        File[] children = dir.listFiles();
        for (int i = 0; i < children.length; i++) {
            if (children[i].isFile()) {
                deleteFile(children[i]);
            } else if (children[i].isDirectory()) {
                deleteDirectory(children[i]);
            }
        }

        boolean success = dir.delete();
        if (success) {
            PersistenceManager persistenceManager = new PersistenceManager();
            persistenceManager.deleteRegistryEntry(getRegsitryKey(dir));
        } else {
            throw new RuntimeException("Cannot delete the resource: " + dir.getName());
        }
    }

    private void createFolder(String parentName, String newFolderName) throws Exception {


        /*
        search for parent. if found, create the new folder in it. this depends on whether we are
        using a remote file system or not.
        add entry 'parentName/newFolderName' as key to the database.
        */

        if (registryType == Constants.LOCAL_HOST_REGISTRY) {
            File parent = new File(localRegistry + parentName);
            if (parent.exists()) {

                File newEntry = new File(parent, newFolderName);

                boolean success = newEntry.mkdir();
                if (!success) {
                    throw new Exception("Couldn't create folder: " + newFolderName);
                }

                // update meta data to the database
                // note that we are using the registry key part (path without the local registry
                // root as the key in db
                // if the new folder has a parent folder, use its values for defaults for all
                // possible properties

                RegistryEntryDO registryEntryDO = new RegistryEntryDO();
                registryEntryDO.setRegistryKey(parentName + URL_SEPARATOR + newFolderName);

                PersistenceManager persistenceManager = new PersistenceManager();
                RegistryEntryDO parentEntryDO =
                        persistenceManager.getRegistryEntry(getURLPath(parentName));
                if (parentEntryDO != null) {
                    registryEntryDO.setExpiryTime(parentEntryDO.getExpiryTime());
                } else {
                    registryEntryDO.setExpiryTime(new Long(getCachableDuration()));
                }

                persistenceManager.addRegistryEntry(registryEntryDO);

            } else {
                throw new SynapseException("Parent folder: " + parentName + " does not exists.");
            }
        }
    }

    private void createFile(String parentName, String newFileName) throws Exception {

        /*
        search for parent. if found, create the new folder in it. this depends on whether we are
        using a remote file system or not.
        */

        if (registryType == Constants.LOCAL_HOST_REGISTRY) {
            File parent = new File(localRegistry + parentName);
            if (parent.exists()) {

                File newFile = new File(parent, newFileName);
                boolean success = newFile.createNewFile();
                if (!success) {
                    throw new Exception("Couldn't create resource: " + newFileName);
                }

                // update meta data to the database
                // note that we are using the registry key part (path without the local registry
                // root as the key in db
                // update the db only if we have some thing else to write other than the key

                RegistryEntryDO registryEntryDO = new RegistryEntryDO();
                registryEntryDO.setRegistryKey(parentName + URL_SEPARATOR + newFileName);

                PersistenceManager persistenceManager = new PersistenceManager();
                RegistryEntryDO parentEntryDO =
                        persistenceManager.getRegistryEntry(getURLPath(parentName));
                if (parentEntryDO != null) {
                    registryEntryDO.setExpiryTime(parentEntryDO.getExpiryTime());
                } else {
                    registryEntryDO.setExpiryTime(new Long(getCachableDuration()));
                }

                persistenceManager.addRegistryEntry(registryEntryDO);

            } else {
                throw new SynapseException("Parent folder: " + parentName + " does not exists.");
            }
        }
    }

    /**
     * Configure the ESB registry using registry parameters.
     *
     * root: file:directory -   registry is on local host
     *                          directory is used to access metadata
     *
     * root: http/https:location -  has to specify one of the following settings
     *                  localRegistry - location of the local registry
     *                  metadataService - url of the service to access metadata
     * If none of above parameters are given "registry" folder is taken as the local registry.
     *
     * @param name
     * @param value
     */
    public void addConfigProperty(String name, String value) {

        if (localRegistry == null) {
            // registry root should always end with "/"
            if (Constants.LOCAL_REGISTRY_ROOT.endsWith(URL_SEPARATOR)) {
                localRegistry = Constants.LOCAL_REGISTRY_ROOT;
            } else {
                localRegistry = Constants.LOCAL_REGISTRY_ROOT + URL_SEPARATOR;
            }
        }

        if(name.equals("root")) {

            // root should always end with '/'
            // therefore, property keys do not have to begin with '/', which could be misleading
            try {
                URL url = new URL(value);

                if (url.getProtocol().equals("file")) {
                    registryProtocol =  FILE;

                    registryType = Constants.LOCAL_HOST_REGISTRY;

                    if (url.getPath().endsWith(URL_SEPARATOR)) {
                        localRegistry = getSystemDependentPath(url.getPath());
                    } else {
                        localRegistry = getSystemDependentPath(url.getPath()) + File.separator;
                    }

                } else if (url.getProtocol().equals("http")) {
                    registryProtocol = HTTP;
                } else if (url.getProtocol().equals("https")) {
                    registryProtocol = HTTPS;
                }


                if(!value.endsWith(URL_SEPARATOR)) {
                    value = value + URL_SEPARATOR;
                }


            } catch (MalformedURLException e) {
                // don't set the root if this is not a valid URL
                throw new RuntimeException("Registry root should be a valid URL.");
            }
        }

        if (name.equals("localRegistry")) {
            registryType = Constants.LOCAL_HOST_REGISTRY;

            // registry root always ends with "/"
            if (!value.endsWith(File.separator)) {
                value = value + File.separator;
            }
            localRegistry = value;
        }

        if (name.equals("matadataService")) {
            registryType = Constants.REMOTE_HOST_REGISTRY;
            metaDataService = value;
        }

        // check if host name verification is disabled
        if (name.equalsIgnoreCase("disableHostNameVerification")) {
            if (value.equalsIgnoreCase("true")) {
                HostnameVerifier hv = new HostnameVerifier()
                {
                    public boolean verify(String urlHostName, SSLSession session)
                    {
                        return true;
                    }
                };

                HttpsURLConnection.setDefaultHostnameVerifier(hv);
            }
        }

        super.addConfigProperty(name, value);
    }

    /**
     * Returns the root of the registry. This is always a URL.
     *
     * @return Registry root.
     */
    public String getRoot() {
        String root = (String) properties.get("root");
        if (root == null) {
            return "";
        } else {
            return root;
        }
    }

    public long getCachableDuration() {
        String cachableDuration = (String) properties.get("cachableDuration");
        return cachableDuration == null ? DEFAULT_CACHABLE_DURATION : Long.parseLong(cachableDuration);
    }

    /**
     * Gives the children of the given entry. If the registry is in the same host get the children
     * (subfolders and files) using the file system. If the registry is in a remote host, get
     * children using a WS call. Give null or registry entry with "" as the key to list the children
     * of the root.
     *
     * @param entry
     * @return children of the given entry
     */
    public RegistryEntry[] getChildren(RegistryEntry entry) {

        String registryRoot = localRegistry;

        if(entry == null) {
            // give the children of the root
            // null or key = "" stands for root

            ESBRegistryEntry urlEntry = new ESBRegistryEntry();
            urlEntry.setKey("");
            entry = urlEntry;
        }

        if (registryType == Constants.LOCAL_HOST_REGISTRY) {

            // registry is in the local file system. access it directly.

            String entryPath = getSystemDependentPath(entry.getKey());
            File file = new File(registryRoot + entryPath);
            if (!file.isDirectory()) {
                return null;
            }

            try {

                String[] children = file.list();
                RegistryEntry[] entries = new RegistryEntry[children.length];
                for (int i = 0; i < children.length; i ++) {

                    ESBRegistryEntry registryEntry = new ESBRegistryEntry();
                    if(entry.getKey().equals("")) {
                        // user asking for the children of the root
                        registryEntry.setKey(children[i]);
                    } else {
                        if(entryPath.endsWith(URL_SEPARATOR)) {
                            registryEntry.setKey(getURLPath(entry.getKey() + children[i]));
                        } else {
                            registryEntry.setKey(getURLPath(entry.getKey() + URL_SEPARATOR + children[i]));
                        }
                    }

                    // set if the registry entry is a file or a folder
                    File entryFile = new File
                            (registryRoot + getSystemDependentPath(registryEntry.getKey()));
                    if (entryFile.isDirectory()) {
                        registryEntry.setType(Constants.folder);
                    }

                    entries[i] = registryEntry;
                }

                return entries;

            } catch(Exception e) {
                throw new SynapseException("Error in reading the URL.");
            }


        } else if (registryType == Constants.REMOTE_HOST_REGISTRY) {
            // implement for remote registries.
        }

        return null;
    }

    public RegistryEntry[] getDescendants(RegistryEntry entry) {

        ArrayList list = new ArrayList();
        RegistryEntry[] entries = getChildren(entry);
        if(entries != null) {
            for(int i=0; i<entries.length; i++) {

                if(list.size() > MAX_KEYS) {
                    break;
                }

                fillDescendants(entries[i], list);
            }
        }

        RegistryEntry[] descendants = new RegistryEntry[list.size()];
        for(int i=0; i<list.size(); i++) {
            descendants[i] = (RegistryEntry) list.get(i);
        }

        return descendants;
    }

    private void fillDescendants(RegistryEntry parent, ArrayList list) {

        RegistryEntry[] entries = getChildren(parent);
        if(entries != null) {
            for(int i=0; i<entries.length; i++) {

                if(list.size() > MAX_KEYS) {
                    break;
                }

                fillDescendants(entries[i], list);
            }
        } else {
            list.add(parent);
        }
    }

    private String getSystemDependentPath(String path) {
        String filePath = path.replace(URL_SEPARATOR_CHAR, File.separatorChar);
        return filePath;
    }

    private String getURLPath(String filePath) {
        return filePath.replace(File.separatorChar, URL_SEPARATOR_CHAR);
    }

    private void handleException(String msg, Exception e) {
        log.error(msg, e);
        throw new SynapseException(msg, e);
    }
}
