/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * 
 *  http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 *  * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.wso2.esb.services;

import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.impl.builder.StAXOMBuilder;
import org.apache.axis2.AxisFault;
import org.apache.axis2.description.Parameter;
import org.apache.axis2.util.XMLPrettyPrinter;
import org.apache.synapse.Constants;
import org.apache.synapse.SynapseException;
import org.apache.synapse.core.axis2.ProxyService;
import org.apache.synapse.core.axis2.MessageContextCreatorForAxis2;
import org.apache.synapse.config.SynapseConfiguration;
import org.apache.synapse.config.xml.XMLConfigurationBuilder;
import org.apache.synapse.config.xml.XMLConfigurationSerializer;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import javax.xml.stream.XMLStreamException;
import java.io.*;
import java.util.Iterator;

/**
 * This is a POJO for the configuration admin service
 */
public class ConfigAdmin extends AbstractESBAdmin {

    private static final Log log = LogFactory.getLog(ConfigAdmin.class);

    public OMElement getConfiguration() throws AxisFault {
        try {
            ByteArrayOutputStream stream = new ByteArrayOutputStream();
            XMLConfigurationSerializer.serializeConfiguration(getSynapseConfiguration(), stream);
            StAXOMBuilder builder
                    = new StAXOMBuilder(new ByteArrayInputStream(stream.toByteArray()));
            return builder.getDocumentElement();
        } catch (XMLStreamException e) {
            log.error("Error serializing the Synapse configuration");
            throw  AxisFault.makeFault(e);
        }
    }

    public void updateConfiguration(OMElement configElement) throws AxisFault {
        try {
            log.info("Stopping Proxy services...");
            Iterator iter = getSynapseConfiguration().getProxyServices().iterator();
            while (iter.hasNext()) {
                ProxyService proxy = (ProxyService) iter.next();
                try {
                    getAxisConfig().removeService(proxy.getName());
                } catch (AxisFault e) {
                    log.error("Error removing Proxy service : " + proxy.getName(), e);
                    throw e;
                }
            }

            ByteArrayOutputStream stream = new ByteArrayOutputStream();
            configElement.getFirstElement().serialize(stream);
            SynapseConfiguration synCfg = XMLConfigurationBuilder.getConfiguration(
                new ByteArrayInputStream(stream.toByteArray()));
            synCfg.setPathToConfigFile(getSynapseConfiguration().getPathToConfigFile());

            Parameter synapseCtxParam = new Parameter(Constants.SYNAPSE_CONFIG, null);
            synapseCtxParam.setValue(synCfg);
            getAxisConfig().addParameter(synapseCtxParam);
            MessageContextCreatorForAxis2.setSynConfig(synCfg);

            // redeploy proxy services
            log.info("Re-deploying Proxy services...");
            iter = synCfg.getProxyServices().iterator();
            while (iter.hasNext()) {
                ProxyService proxy = (ProxyService) iter.next();
                proxy.buildAxisService(synCfg, getAxisConfig());
                log.info("Deployed Proxy service : " + proxy.getName());
                if (!proxy.isStartOnLoad()) {
                    proxy.stop(synCfg);
                }
            }
            log.info("Configuration updated successfully...");

        } catch (SynapseException se) {
            handleFault(log, "Unable to update the Synapse configuration." +
                se.getMessage() +" Check log for more details", se);
        } catch (XMLStreamException e) {
            handleFault(log, "Unable to update the Synapse configuration." +
                e.getMessage() +" Check log for more details", e);
        }
    }

    public boolean saveConfigurationToDisk() throws AxisFault {
        log.info("Saving configuration..");

        SynapseConfiguration config = getSynapseConfiguration();
        try {
            FileOutputStream fos = new FileOutputStream(config.getPathToConfigFile());
            XMLConfigurationSerializer.serializeConfiguration(config, fos);
            try {
                fos.close();
                XMLPrettyPrinter.prettify(new File(config.getPathToConfigFile()));
            }
            catch (IOException e) {
                // ignore prettyfy errors
            }
            log.info("Configuration saved to disk");
            return true;
        }
        catch (XMLStreamException e) {
            handleFault(log, "Could not save changes to disk." +
                e.getMessage() + " Check log for more details", e);
        }
        catch (FileNotFoundException e) {
            handleFault(log, "Could not locate the Synapse configuration file to save changes", e);
        }
        return false;
    }

}
