/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.esb.services;

import org.apache.axiom.om.OMElement;
import org.apache.axis2.AxisFault;
import org.apache.synapse.SynapseException;
import org.apache.synapse.config.SynapseConfiguration;
import org.apache.synapse.config.xml.MediatorFactoryFinder;
import org.apache.synapse.config.xml.MediatorSerializerFinder;
import org.apache.synapse.config.xml.XMLConfigurationBuilder;
import org.apache.synapse.mediators.base.SequenceMediator;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.esb.services.tos.SequenceData;

import javax.xml.namespace.QName;
import java.util.*;

/**
 * This is a POJO for the Sequence admin service
 */
public class SequenceAdmin extends AbstractESBAdmin {

    private static final Log log = LogFactory.getLog(SequenceAdmin.class);

    /**
     * Gets the details about the sequences as a SequenceData array
     *
     * @return Array of SequenceData representing the Sequences in the
     *         SynapseConfiguration
     * @throws AxisFault if an error occured while getting the data from
     *                   SynapseConfiguration
     */
    public SequenceData[] sequenceData() throws AxisFault {
        SynapseConfiguration synapseConfiguration = getSynapseConfiguration();
        Map definedsequencesMap = synapseConfiguration.getDefinedSequences();
        Collection definedsequenceValues = definedsequencesMap.values();
        ArrayList sequenceDataList = new ArrayList();
        for (Iterator ite = definedsequenceValues.iterator(); ite.hasNext();) {
            Object obj = ite.next();
            /*
            instanceof will have SequenceMediator || GolableConfigurations
            */
            if (obj instanceof SequenceMediator) {
                SequenceMediator seqMediator = (SequenceMediator) obj;
                SequenceData data = new SequenceData();
                data.setName(seqMediator.getName());
                if(seqMediator.getStatisticsEnable() == org.apache.synapse.Constants.STATISTICS_ON) {
                    data.setEnableStatistics(true);
                } else {
                    data.setEnableStatistics(false);
                }
                if(seqMediator.getTraceState() == org.apache.synapse.Constants.TRACING_ON) {
                    data.setEnableTracing(true);
                } else {
                    data.setEnableTracing(false);
                }
                sequenceDataList.add(data);
            }
        }

        Collections.sort(sequenceDataList, new Comparator() {
            public int compare(Object o1, Object o2) {
                return ((SequenceData) o1).getName().compareTo(
                    ((SequenceData)o2).getName());
            }
        });

        return (SequenceData[]) sequenceDataList
                .toArray(new SequenceData[sequenceDataList.size()]);
    }

    /**
     * Deletes the sequence with the given name from SynapseConfiguration
     *
     * @param sequenceName - Name of the sequence to delete
     * @throws AxisFault if the Sequence described by the given name doesnt
     *                   exists in the Synapse Configuration
     */
    public void deleteSequence(String sequenceName) throws AxisFault {
        try {
            log.debug("Deleting sequence : " + sequenceName);
            if("main".equals(sequenceName) || "fault".equals(sequenceName)) {
                handleFault(log, "Unable to delete sequence : " + sequenceName, null);
            } else {
                SynapseConfiguration synapseConfiguration
                        = getSynapseConfiguration();
                synapseConfiguration.removeSequence(sequenceName);
                log.info("Deleted sequence : " + sequenceName + " from the configuration");
            }
        } catch (SynapseException syne) {
            handleFault(log, "Error removing sequence : " + sequenceName, syne);
        }
    }

    /**
     * todo - docs
     * @param sequenceName
     * @throws AxisFault
     */
    public void enableStatistics(String sequenceName) throws AxisFault {
        try {
            ((SequenceMediator)getSynapseConfiguration().getSequence(
                    sequenceName)).setStatisticsEnable(org.apache.synapse.Constants.STATISTICS_ON);
            log.info("Enabled statistics on sequence : " + sequenceName);
        } catch (SynapseException syne) {
            handleFault(log,
                    "Unable to enable statistics for the sequence " + sequenceName, syne);
        }
    }

    /**
     * todo - docs
     * @param sequenceName
     * @throws AxisFault
     */
    public void disableStatistics(String sequenceName) throws AxisFault {
        try {
            ((SequenceMediator)getSynapseConfiguration().getSequence(
                    sequenceName)).setStatisticsEnable(org.apache.synapse.Constants.STATISTICS_OFF);
            log.info("Disabled statistics on sequence : " + sequenceName);
        } catch (SynapseException syne) {
            handleFault(log,
                    "Unable to disable statistics for the sequence " + sequenceName, syne);
        }
    }

    /**
     * 
     * @param sequenceName
     * @throws AxisFault
     */
    public void enableTracing(String sequenceName) throws AxisFault {

        try {
            ((SequenceMediator)getSynapseConfiguration().getSequence(
                    sequenceName)).setTraceState(org.apache.synapse.Constants.TRACING_ON);
            log.info("Enabled tracing on sequence : " + sequenceName);
        } catch (SynapseException syne) {
            handleFault(log, "Unable to enable tracing for the sequence " + sequenceName, syne);
        }
    }

    /**
     * 
     * @param sequenceName
     * @throws AxisFault
     */
    public void disableTracing(String sequenceName) throws AxisFault {

        try {
            ((SequenceMediator)getSynapseConfiguration().getSequence(
                    sequenceName)).setTraceState(org.apache.synapse.Constants.TRACING_OFF);
            log.info("Disabled tracing on sequence : " + sequenceName);
        } catch (SynapseException syne) {
            handleFault(log, "Unable to disable tracing for the sequence " + sequenceName, syne);
        }
    }

    /**
     * Add a sequence into the synapseConfiguration
     *
     * @param sequenceElement - Sequence object to be added as an OMElement
     * @throws AxisFault if a sequence exists with the same name or if the
     *                   element provided is not a Sequence element
     */
    public void addSequence(OMElement sequenceElement) throws AxisFault {

        try {
            if (sequenceElement.getFirstElement().getQName().getLocalPart()
                    .equals(org.apache.synapse.config.xml.Constants
                            .SEQUENCE_ELT.getLocalPart())) {
                String sequenceName = sequenceElement.getFirstElement().getAttributeValue(
                        new QName("name"));

                log.info("Adding sequence : " + sequenceName + " to the configuration");
                if (getSynapseConfiguration().getLocalRegistry().get(sequenceName) != null) {
                    handleFault(log, "The name '" + sequenceName +
                        "' is already used within the configuration", null);
                } else {
                    XMLConfigurationBuilder.defineSequence(getSynapseConfiguration(),
                            sequenceElement.getFirstElement());
                    log.info("Added sequence : " + sequenceName + " to the configuration");
                }
            } else {
                handleFault(log, "Error adding sequence. Invalid definition", null);
            }
        } catch (SynapseException syne) {
            handleFault(log, "Error adding sequence", syne);
        }
    }

    /**
     * Returns the OMelement representation of the sequence given by sequence
     * name
     *
     * @param sequenceName - name of the sequence to get
     * @return OMElement representing the SequenceMediator of the given sequence
     *         name
     * @throws AxisFault if any error occured while getting the data from the
     *                   SynapseConfiguration
     */
    public OMElement getSequence(String sequenceName) throws AxisFault {
        SynapseConfiguration synapseConfiguration = getSynapseConfiguration();
        try {
            if (synapseConfiguration.getSequence(sequenceName) != null) {
                return MediatorSerializerFinder.getInstance().getSerializer(
                        synapseConfiguration.getSequence(sequenceName))
                        .serializeMediator(null, synapseConfiguration
                                .getSequence(sequenceName));
            } else {
                handleFault(log, "Sequence with the name " + sequenceName + " does not exist", null);
            }
        } catch (SynapseException syne) {
            handleFault(log, "Unable to get the sequence : " + sequenceName, syne);
        }
        return null;
    }

    /**
     * Returns the OMElement representation of the main mediator sequence
     *
     * @return OMElement representing the main mediator sequence
     * @throws AxisFault if an error occurs while retrieving the main mediator from Synapse
     */
    public OMElement getMainSequence() throws AxisFault {
        SynapseConfiguration synapseConfiguration = getSynapseConfiguration();
        try {
            if (synapseConfiguration.getMainSequence() != null) {
                return MediatorSerializerFinder.getInstance().getSerializer(
                        synapseConfiguration.getMainSequence())
                        .serializeMediator(null, synapseConfiguration.getMainSequence());
            } else {
                handleFault(log, "The main mediator sequence " +
                        "does not exist in the configuration", null);
            }
        } catch (SynapseException syne) {
            handleFault(log, "Unable to get or serialize the main mediator sequence", syne);
        }
        return null;
    }

    /**
     * Saves the sequence described with the sequence string
     *
     * @param sequenceElement - String representing the XML describing the
     *                        Sequence
     * @throws AxisFault if the sequence name already exists or if the string
     *                   doesnt represent a Sequence element
     */
    public void saveSequence(OMElement sequenceElement) throws AxisFault {
        try {
            sequenceElement = sequenceElement.getFirstChildWithName(
                    org.apache.synapse.config.xml.Constants.SEQUENCE_ELT);
            if (sequenceElement != null) {
                String sequenceName = sequenceElement.getAttributeValue(new QName("name"));

                log.debug("Saving sequence : " + sequenceName);
                if (getSynapseConfiguration().getSequence(sequenceName)
                        == null) {
                    handleFault(log, "Unable to sa ve sequence " +
                        sequenceName + ". Does not exist", null);
                } else {
                    getSynapseConfiguration().addSequence(sequenceName,
                            MediatorFactoryFinder.getInstance().getMediator(sequenceElement));
                    log.debug("Saved sequence : " + sequenceName + " to the configuration");
                }
            } else {
                handleFault(log, "Unable to save sequence. Invalid definition", null);
            }
        } catch (SynapseException syne) {
            handleFault(log, "Unable to save the sequence", syne);
        }
    }

}
