/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.esb.transport.jetty;

import org.apache.axiom.om.OMElement;
import org.apache.axis2.AxisFault;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.description.TransportInDescription;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.mortbay.jetty.Connector;
import org.mortbay.jetty.Handler;
import org.mortbay.jetty.NCSARequestLog;
import org.mortbay.jetty.Server;
import org.mortbay.jetty.servlet.SessionHandler;
import org.mortbay.jetty.servlet.HashSessionIdManager;
import org.mortbay.jetty.handler.*;
import org.mortbay.jetty.security.SslSocketConnector;
import org.wso2.utils.NetworkUtils;
import org.wso2.utils.transport.http.HttpServerManager;
import org.wso2.esb.util.JettyWebConfiguration;

import javax.xml.namespace.QName;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.net.URL;

public class JettyServer {

    private static Log log = LogFactory.getLog(JettyServer.class);
    private String adminWebRoot = "./www/";

    protected ConfigurationContext configurationContext;
    private static JettyWebConfiguration jettyConfig = JettyWebConfiguration.getInstance();

    public JettyServer(ConfigurationContext systemContext) {
        configurationContext = systemContext;
    }

    /**
     * Start this server as a NON-daemon.
     */
    public void start() throws AxisFault {
        try {
            Server adminServer = HttpServerManager.getServer(Constants.ADMIN_SERVER);
            ContextHandlerCollection adminContext = new ContextHandlerCollection();
            HandlerCollection adminHandlers = new HandlerCollection();

            ContextHandler soapCtx = new ContextHandler();
            soapCtx.setContextPath(initializeServiceUrlContextPath(configurationContext));
            soapCtx.addHandler(new Axis2SOAPHandler(configurationContext));

            ContextHandler contextRootSoap = new ContextHandler();
            contextRootSoap.setContextPath("/");
            contextRootSoap.addHandler(new Axis2SOAPHandler(configurationContext));

            ContextHandler contextRootRest = new ContextHandler();
            contextRootRest.setContextPath("/");
            contextRootRest.addHandler(new Axis2RestHandler(configurationContext));

            ContextHandler adminCtx = new ContextHandler();
            adminCtx.setContextPath(jettyConfig.adminWebCtx);
            adminCtx.setResourceBase(jettyConfig.adminWebResourceBase);
            adminCtx.addHandler(new ManagementConsoleRequestHandler(configurationContext));

            ContextHandler stylesCtx = new ContextHandler();
            stylesCtx.setContextPath(jettyConfig.stylesWebCtx);
            ResourceHandler schemaWebResourceHandler = new ResourceHandler();
            schemaWebResourceHandler.setResourceBase(jettyConfig.stylesWebResourceBase);
            stylesCtx.addHandler(schemaWebResourceHandler);

            ContextHandler docsCtx = new ContextHandler();
            docsCtx.setContextPath(jettyConfig.docsWebCtx);
            ResourceHandler docsResourceHandler = new ResourceHandler();
            docsResourceHandler.setResourceBase(jettyConfig.docsWebResourceBase);
            docsCtx.addHandler(docsResourceHandler);
            
//            RequestLogHandler requestLogHandler = new RequestLogHandler();
//            NCSARequestLog requestLog = new NCSARequestLog(
//                "./logs/jetty-" + (new SimpleDateFormat("yyyy-MM-dd").format(new Date())) + ".log");
//            requestLog.setExtended(false);
//            requestLog.setAppend(true);
//            requestLog.setRetainDays(1);
//            requestLog.setLogTimeZone("GMT");
//            requestLogHandler.setRequestLog(requestLog);

            ContextHandler registryCtx = new ContextHandler();
            registryCtx.setContextPath("/registry");
            ResourceHandler registryResourceHandler = new ResourceHandler();
            registryResourceHandler.setResourceBase("./registry");
            SessionHandler registrySessionHandler = new SessionHandler();
            registrySessionHandler.addHandler(registryResourceHandler);
            registryCtx.addHandler(registrySessionHandler);

            adminContext.setHandlers(new Handler[] {
                stylesCtx, adminCtx, docsCtx, soapCtx, contextRootSoap, registryCtx});
            adminHandlers.setHandlers(
                    new Handler[]{adminContext, new ServerDefaultHandler()/*, requestLogHandler*/});
            adminServer.setHandler(adminHandlers);

            startAdminServer();

        } catch (Exception e) {
            String msg = "Cannot start Jetty server";
            log.error(msg, e);
            throw new AxisFault(msg, e);
        }
    }

    private void startAdminServer() throws Exception {

        TransportInDescription tInDescription = configurationContext.getAxisConfiguration()
            .getTransportIn("https");

        Server adminServer = HttpServerManager.getServer(Constants.ADMIN_SERVER);
        SslSocketConnector sslSocketConnector = new SslSocketConnector();
        sslSocketConnector.setPort(9443);
        int maxIdleTimeMs = 120000; // 2 mins
        OMElement ksEle =
                tInDescription.getParameter("keystore").getParameterElement()
                        .getFirstElement();
        String ksLocation =
                ksEle.getFirstChildWithName(new QName("Location")).getText();
        String ksType =
                ksEle.getFirstChildWithName(new QName("Type")).getText();
        String ksPassword =
                ksEle.getFirstChildWithName(new QName("Password")).getText();
        String pvtKeyPassword =
                ksEle.getFirstChildWithName(new QName("KeyPassword")).getText();

        sslSocketConnector.setMaxIdleTime(maxIdleTimeMs);
        URL url = getClass().getClassLoader().getResource(ksLocation);
        sslSocketConnector.setKeystore(url.getPath());
        sslSocketConnector.setKeystoreType(ksType);
        sslSocketConnector.setPassword(ksPassword);
        sslSocketConnector.setKeyPassword(pvtKeyPassword);

        adminServer.setConnectors(new Connector[]{sslSocketConnector});
        adminServer.start();

        log.info("Console listener started on port 9443 over HTTPS. " +
            "Connect to https://" + NetworkUtils.getLocalHostname() + ":9443/");
    }


    /**
     * Stop this server. Can be called safely if the system is already stopped,
     * or if it was never started. This will interrupt any pending accept().
     */
    public void stop() throws AxisFault {
        log.info("Stopping Jetty server...");
        try {
            HttpServerManager.stopAllServers();
        } catch (Exception e) {
            log.error("InterruptedException while trying to stop HTTP servers.");
            throw  AxisFault.makeFault(e);
        }
        log.info("Jetty server stopped.");
    }

    /**
     * Method getSystemContext
     *
     * @return the system context
     */
    public ConfigurationContext getSystemContext() {
        return configurationContext;
    }

    public ConfigurationContext getConfigurationContext() {
        return configurationContext;
    }

    public String getAdminWebRoot() {
        return adminWebRoot;
    }

    public void setAdminWebRoot(String adminWebRoot) {
        this.adminWebRoot = adminWebRoot;
    }

    private String initializeServiceUrlContextPath(ConfigurationContext configurationContext) {

        return configurationContext.getServiceContextPath();
    }

//    private String initializeServiceUrlRestPath(ConfigurationContext cfgCtx) {
//        return cfgCtx.ggetRESTContextPath();
//    }
}
